from transphone.g2p import read_g2p
from transphone.bin.download_model import download_model
from transphone.model.utils import get_all_models, resolve_model_name
from pathlib import Path
import argparse

if __name__ == '__main__':

    parser = argparse.ArgumentParser('running transphone g2p model')
    parser.add_argument('-m', '--model', type=str, default='latest', help='specify which model to use. default is to use the latest local model')
    parser.add_argument('-l', '--lang', type=str,  default='ipa',help='specify which language inventory to use for recognition. default is to use all phone inventory')
    parser.add_argument('-i', '--input', type=str, required=True, help='specify your input wav file/directory')
    parser.add_argument('-o', '--output', type=str, default='stdout', help='specify output file. the default will be stdout')
    parser.add_argument('-c', '--combine', type=bool, default=False, help='write outputs by including both grapheme inputs and phonemes in the same line, delimited by space')

    args = parser.parse_args()

    # download specified model automatically if no model exists
    if len(get_all_models()) == 0:
        download_model('latest')

    # resolve model's name
    model_name = resolve_model_name(args.model)
    if model_name == "none":
        print("Model ", model_name, " does not exist. Please download this model or use an existing model in list_model")
        exit(0)

    args.model = model_name

    # create model
    g2p = read_g2p(model_name)

    # output file descriptor
    output_fd = None
    if args.output != 'stdout':
        output_fd = open(args.output, 'w', encoding='utf-8')

    # cache infered words
    word_cache = {}

    # input file/path
    input_path = Path(args.input)

    for line in open(input_path, 'r'):
        words = line.strip().split()
        phoneme_lst = []

        for word in words:
            if word in word_cache:
                phonemes = word_cache[word]
            else:
                phonemes = g2p.inference(word, args.lang)

            phoneme_lst.extend(phonemes)

        if args.combine:
            line_output = ' '.join(words)+' '+' '.join(phoneme_lst)

        else:
            line_output = ' '.join(phoneme_lst)

        if output_fd:
            output_fd.write(line_output+'\n')
        else:
            print(line_output)


    if output_fd:
        output_fd.close()


from phonepiece.unit import Unit
from pathlib import Path
from phonepiece.config import phonepiece_config
from unidecode import unidecode
import editdistance

def read_grapheme(lang_id):

    unit_to_id = dict()

    unit_to_id['<blk>'] = 0

    idx = 0

    unit_path = Path(phonepiece_config.data_path / 'phonetisaurus' / lang_id / 'char.txt')

    for line in open(str(unit_path), 'r', encoding='utf-8'):
        fields = line.strip().split()

        assert len(fields) < 3

        if len(fields) == 1:
            unit = fields[0]
            idx += 1
        else:
            unit = fields[0]
            idx = int(fields[1])

        unit_to_id[unit] = idx

    unit_to_id['<blk>'] = 0
    unit_to_id['<eos>'] = idx+1

    return Grapheme(unit_to_id)


class Grapheme(Unit):

    def __init__(self, unit_to_id):
        super().__init__(unit_to_id)

        self.latins = []
        self.nearest_mapping = None

        for i, elem in enumerate(self.elems):
            if i == 0:
                self.latins.append('<blk>')
            else:
                self.latins.append(unidecode(elem))

    def get_nearest_unit(self, unit):

        if self.nearest_mapping is None:
            self.nearest_mapping = dict()

        if unit in self.nearest_mapping:
            return self.nearest_mapping[unit]

        if unit in self.unit_to_id:
            self.nearest_mapping[unit] = unit
            return unit

        target_latin = unidecode(unit)

        edit_score = dict()
        for i, latin in enumerate(self.latins):
            unit = self.id_to_unit[i]
            edit_score[unit] = editdistance.eval(latin, target_latin)

        target_unit = min(edit_score, key=edit_score.get)
        self.nearest_mapping[unit] = target_unit

        return target_unit


from transphone.config import TransphoneConfig
from transphone.data.vocab import Vocab
from tqdm import tqdm
import torch
import random

def read_dataset(lang_id='eng'):

    r = open(TransphoneConfig.lang_path / lang_id / 'new_input', 'r')

    phoneme_lst = []
    grapheme_lst = []

    phoneme_set = set()
    grapheme_set = set()

    for line in r:
        phoneme_str, grapheme_str = line.strip().split('\t')
        phonemes = phoneme_str.split()
        graphemes = grapheme_str.split()
        phoneme_lst.append(phonemes)
        grapheme_lst.append(graphemes)
        phoneme_set.update(phonemes)
        grapheme_set.update(graphemes)

    phoneme_vocab = Vocab(phoneme_set)
    grapheme_vocab = Vocab(grapheme_set)

    return Dataset(phoneme_lst, grapheme_lst, phoneme_vocab, grapheme_vocab)


def read_p2g_dataset():
    r = open('/home/xinjianl/Git/allospeech/data/wilderness/bfa/exp/train.txt', 'r')

    train_phoneme_lst = []
    train_grapheme_lst = []

    test_phoneme_lst = []
    test_grapheme_lst = []

    phoneme_set = set()
    grapheme_set = set()

    max_len = 128
    for line in r:
        phoneme_str, grapheme_str = line.strip().split('\t')
        phonemes = phoneme_str.split()
        graphemes = grapheme_str.split()

        if len(phonemes) > max_len or len(graphemes) > max_len:
            continue

        train_phoneme_lst.append(phonemes)
        train_grapheme_lst.append(graphemes)
        phoneme_set.update(phonemes)
        grapheme_set.update(graphemes)


    r.close()

    r = open('/home/xinjianl/Git/allospeech/data/wilderness/bfa/exp/test.txt', 'r')
    for line in r:
        phoneme_str, grapheme_str = line.strip().split('\t')
        phonemes = phoneme_str.split()
        graphemes = grapheme_str.split()

        if len(phonemes) > max_len or len(graphemes) > max_len:
            continue

        test_phoneme_lst.append(phonemes)
        test_grapheme_lst.append(graphemes)
        phoneme_set.update(phonemes)
        grapheme_set.update(graphemes)

    phoneme_vocab = Vocab(phoneme_set)
    grapheme_vocab = Vocab(grapheme_set)

    train_set = P2GDataset(train_phoneme_lst, train_grapheme_lst, phoneme_vocab, grapheme_vocab)
    test_set = P2GDataset(test_phoneme_lst, test_grapheme_lst, phoneme_vocab, grapheme_vocab)

    return train_set, test_set

def read_multilingual_dataset():

    phoneme_lst = []
    grapheme_lst = []

    test_phoneme_lst = []
    test_grapheme_lst = []

    phoneme_set = set()
    grapheme_set = set()

    print("loading dataset...")

    for lang_dir in tqdm(list(TransphoneConfig.lang_path.glob('*'))):
        lang_id = lang_dir.stem

        if not (lang_dir / 'new_input').exists():
            continue
        lang_tag = '<'+lang_id+'>'

        grapheme_set.add(lang_tag)

        r = open(lang_dir / 'new_input', 'r')

        lines = r.readlines()

        random.shuffle(lines)

        # ignore the last 25
        max_num = max(0, len(lines)-5)
        min_num = min(max_num, 10000)

        for line in lines[:min_num]:
            grapheme_str, phoneme_str = line.strip().split('\t')
            grapheme_str = lang_tag + ' '+grapheme_str
            phonemes = phoneme_str.split()
            graphemes = grapheme_str.split()
            phoneme_lst.append(phonemes)
            grapheme_lst.append(graphemes)
            phoneme_set.update(phonemes)
            grapheme_set.update(graphemes)

        # last 25 for test
        for line in lines[-5:]:
            grapheme_str, phoneme_str = line.strip().split('\t')
            grapheme_str = lang_tag + ' '+ grapheme_str
            phonemes = phoneme_str.split()
            graphemes = grapheme_str.split()
            test_phoneme_lst.append(phonemes)
            test_grapheme_lst.append(graphemes)
            phoneme_set.update(phonemes)
            grapheme_set.update(graphemes)

    phoneme_vocab = Vocab(phoneme_set)
    grapheme_vocab = Vocab(grapheme_set)

    train_dataset = Dataset(phoneme_lst, grapheme_lst, phoneme_vocab, grapheme_vocab)
    test_dataset = Dataset(test_phoneme_lst, test_grapheme_lst, phoneme_vocab, grapheme_vocab)

    return train_dataset, test_dataset


class Dataset:

    def __init__(self, phoneme_lst, grapheme_lst, phoneme_vocab, grapheme_vocab):

        self.phoneme_lst = phoneme_lst
        self.grapheme_lst = grapheme_lst
        self.phoneme_vocab = phoneme_vocab
        self.grapheme_vocab = grapheme_vocab

    def __getitem__(self, item):

        phones = self.phoneme_lst[item]
        graphemes = self.grapheme_lst[item]

        phone_ids = [self.phoneme_vocab.atoi(phone) for phone in phones]
        grapheme_ids = [self.grapheme_vocab.atoi(grapheme) for grapheme in graphemes]

        return (torch.LongTensor(grapheme_ids), torch.LongTensor(phone_ids))

    def __len__(self):
        return len(self.phoneme_lst)


class P2GDataset:

    def __init__(self, phoneme_lst, grapheme_lst, phoneme_vocab, grapheme_vocab):

        self.phoneme_lst = phoneme_lst
        self.grapheme_lst = grapheme_lst
        self.phoneme_vocab = phoneme_vocab
        self.grapheme_vocab = grapheme_vocab

    def __getitem__(self, item):

        phones = self.phoneme_lst[item]
        graphemes = self.grapheme_lst[item]

        phone_ids = [self.phoneme_vocab.atoi(phone) for phone in phones]
        grapheme_ids = [self.grapheme_vocab.atoi(grapheme) for grapheme in graphemes]

        return (torch.LongTensor(phone_ids), torch.LongTensor(grapheme_ids))

    def __len__(self):
        return len(self.phoneme_lst)


from transphone.model.checkpoint_utils import torch_load
from transphone.model.transformer import TransformerG2P
from transphone.model.ensemble import ensemble
from transphone.bin.download_model import download_model
from transphone.config import TransphoneConfig
from transphone.data.vocab import Vocab
from phonepiece.tree import read_tree
from phonepiece.inventory import read_inventory
import torch


def read_g2p(model_name='latest', alt_model_path=None):

    if alt_model_path:
        # check whether a customized path is used or not
        model_path = alt_model_path
    else:
        model_path = TransphoneConfig.data_path / 'model' / model_name

    # if not exists, we try to download the model
    if not model_path.exists():
        download_model(model_name)

    if not model_path.exists():
        raise ValueError(f"could not download or read {model_name} inventory")

    model = G2P(model_path, {})

    return model


class G2P:

    def __init__(self, model_path, inference_config):

        self.model_path = model_path
        self.grapheme_vocab = Vocab.read(model_path / 'grapheme.vocab')
        self.phoneme_vocab = Vocab.read(model_path / 'phoneme.vocab')
        self.inference_config = inference_config

        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

        # setup available languages
        self.supervised_langs = []
        for word in self.grapheme_vocab.words[2:]:
            if len(word) == 5 and word[0] == '<' and word[-1] == '>':
                self.supervised_langs.append(word[1:-1])


        # cache to find proper supervised language
        self.lang_map = {}

        # inventory
        self.lang2inv = {}

        # tree to estimate language's similarity
        self.tree = read_tree()
        self.tree.setup_target_langs(self.supervised_langs)
        self.supervised_langs = set(self.supervised_langs)


        SRC_VOCAB_SIZE = len(self.grapheme_vocab)+1
        TGT_VOCAB_SIZE = len(self.phoneme_vocab)+1

        EMB_SIZE = 512
        NHEAD = 8
        FFN_HID_DIM = 512
        NUM_ENCODER_LAYERS = 4
        NUM_DECODER_LAYERS = 4
        torch.manual_seed(0)


        self.model = TransformerG2P(NUM_ENCODER_LAYERS, NUM_DECODER_LAYERS, EMB_SIZE,
                            NHEAD, SRC_VOCAB_SIZE, TGT_VOCAB_SIZE, FFN_HID_DIM).to(self.device)


        torch_load(self.model, model_path / "model.pt")


    def get_target_langs(self, lang_id, num_lang=10, force_approximate=False):

        if lang_id in self.lang_map:
            target_langs = self.lang_map[lang_id]
        else:

            if force_approximate or lang_id not in self.supervised_langs:
                target_langs = self.tree.get_nearest_langs(lang_id, num_lang)
                print("lang ", lang_id, " is not available directly, use ", target_langs, " instead")
                self.lang_map[lang_id] = target_langs
            else:
                self.lang_map[lang_id] = [lang_id]
                target_langs = [lang_id]

        return target_langs

    def inference(self, word, lang_id='eng', num_lang=10, debug=False, force_approximate=False):

        target_langs = self.get_target_langs(lang_id, num_lang, force_approximate)

        phones_lst = []

        for target_lang_id in target_langs:
            lang_tag = '<' + target_lang_id + '>'

            graphemes = [lang_tag]+[w.lower() for w in list(word)]

            grapheme_ids = []
            for grapheme in graphemes:
                if grapheme not in self.grapheme_vocab:
                    print("WARNING: not found grapheme ", grapheme, " in vocab")
                    continue
                grapheme_ids.append(self.grapheme_vocab.atoi(grapheme))

            x = torch.LongTensor(grapheme_ids).unsqueeze(0).to(self.device)

            phone_ids = self.model.inference(x)

            phones = [self.phoneme_vocab.itoa(phone) for phone in phone_ids]

            # ignore empty
            if len(phones) == 0:
                continue

            # if it is a mapped language, we need to map the inference_phone to the correct language inventory
            if lang_id not in self.lang2inv:
                inv = read_inventory(lang_id)
                self.lang2inv[lang_id] = inv

            inv = self.lang2inv[lang_id]
            phones = inv.remap(phones)

            if debug:
                print(target_lang_id, ' ', phones)

            phones_lst.append(phones)


        if len(phones_lst) == 0:
            phones = []
        else:
            phones = ensemble(phones_lst)

        return phones


import torch

def pad_list(tensor_lst):
    max_length = max(t.size(0) for t in tensor_lst)
    batch_size = len(tensor_lst)

    padded_tensor = tensor_lst[0].new(batch_size, max_length, *tensor_lst[0].size()[1:]).fill_(0)
    for i, t in enumerate(tensor_lst):
        padded_tensor[i,:t.size(0)] = t

    return padded_tensor


def pad_sos_eos(ys, sos, eos):

    batch_size = len(ys)

    sos_tensor = ys.new_zeros((batch_size, 1)).fill_(sos)
    y_in = torch.cat([sos_tensor, ys], dim=1)

    y_out = []

    zero_tensor = ys.new_zeros((batch_size, 1))
    extended_ys = torch.cat([ys, zero_tensor], dim=1)

    for y in extended_ys:
        #eos_idx = (y==0).nonzero()[0].item()
        #print(y)
        eos_idx = torch.nonzero(y==0)[0].item()
        y[eos_idx] = eos
        y_out.append(y)

    y_out = pad_list(y_out)
    return y_in, y_out


from torch import Tensor
import torch
import torch.nn as nn
from torch.nn import Transformer
import math
from transphone.data.utils import pad_sos_eos

DEVICE = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

UNK_IDX, PAD_IDX, BOS_IDX, EOS_IDX = 0, 0, 1, 1

def generate_square_subsequent_mask(sz):
    mask = (torch.triu(torch.ones((sz, sz), device=DEVICE)) == 1).transpose(0, 1)
    mask = mask.float().masked_fill(mask == 0, float('-inf')).masked_fill(mask == 1, float(0.0))
    return mask


def create_mask(src, tgt):
    src_seq_len = src.shape[0]
    tgt_seq_len = tgt.shape[0]

    tgt_mask = generate_square_subsequent_mask(tgt_seq_len)
    src_mask = torch.zeros((src_seq_len, src_seq_len),device=DEVICE).type(torch.bool)

    src_padding_mask = (src == PAD_IDX).transpose(0, 1)
    tgt_padding_mask = (tgt == PAD_IDX).transpose(0, 1)
    return src_mask, tgt_mask, src_padding_mask, tgt_padding_mask


# helper Module that adds positional encoding to the token embedding to introduce a notion of word order.
class PositionalEncoding(nn.Module):
    def __init__(self,
                 emb_size: int,
                 dropout: float,
                 maxlen: int = 5000):
        super(PositionalEncoding, self).__init__()
        den = torch.exp(- torch.arange(0, emb_size, 2)* math.log(10000) / emb_size)
        pos = torch.arange(0, maxlen).reshape(maxlen, 1)
        pos_embedding = torch.zeros((maxlen, emb_size))
        pos_embedding[:, 0::2] = torch.sin(pos * den)
        pos_embedding[:, 1::2] = torch.cos(pos * den)
        pos_embedding = pos_embedding.unsqueeze(-2)

        self.dropout = nn.Dropout(dropout)
        self.register_buffer('pos_embedding', pos_embedding)

    def forward(self, token_embedding: Tensor):
        return self.dropout(token_embedding + self.pos_embedding[:token_embedding.size(0), :])

# helper Module to convert tensor of input indices into corresponding tensor of token embeddings
class TokenEmbedding(nn.Module):
    def __init__(self, vocab_size: int, emb_size):
        super(TokenEmbedding, self).__init__()
        self.embedding = nn.Embedding(vocab_size, emb_size)
        self.emb_size = emb_size

    def forward(self, tokens: Tensor):
        return self.embedding(tokens.long()) * math.sqrt(self.emb_size)


# Seq2Seq Network
class TransformerG2P(nn.Module):
    def __init__(self,
                 num_encoder_layers: int,
                 num_decoder_layers: int,
                 emb_size: int,
                 nhead: int,
                 src_vocab_size: int,
                 tgt_vocab_size: int,
                 dim_feedforward: int = 512,
                 dropout: float = 0.1):
        super(TransformerG2P, self).__init__()
        self.transformer = Transformer(d_model=emb_size,
                                       nhead=nhead,
                                       num_encoder_layers=num_encoder_layers,
                                       num_decoder_layers=num_decoder_layers,
                                       dim_feedforward=dim_feedforward,
                                       dropout=dropout)
        self.generator = nn.Linear(emb_size, tgt_vocab_size)
        self.src_tok_emb = TokenEmbedding(src_vocab_size, emb_size)
        self.tgt_tok_emb = TokenEmbedding(tgt_vocab_size, emb_size)
        self.positional_encoding = PositionalEncoding(
            emb_size, dropout=dropout)

        self.criterion = torch.nn.CrossEntropyLoss(ignore_index=PAD_IDX)


    def forward(self,
                src: Tensor,
                trg: Tensor,
                src_mask: Tensor,
                tgt_mask: Tensor,
                src_padding_mask: Tensor,
                tgt_padding_mask: Tensor,
                memory_key_padding_mask: Tensor):
        src_emb = self.positional_encoding(self.src_tok_emb(src))
        tgt_emb = self.positional_encoding(self.tgt_tok_emb(trg))
        outs = self.transformer(src_emb, tgt_emb, src_mask, tgt_mask, None,
                                src_padding_mask, tgt_padding_mask, memory_key_padding_mask)
        return self.generator(outs)

    def encode(self, src: Tensor, src_mask: Tensor):
        return self.transformer.encoder(self.positional_encoding(self.src_tok_emb(src)), src_mask)

    def decode(self, tgt: Tensor, memory: Tensor, tgt_mask: Tensor):
        return self.transformer.decoder(self.positional_encoding(self.tgt_tok_emb(tgt)), memory,tgt_mask)


    def train_step(self, x,y):

        self.train()
        batch_size = x.shape[0]

        ys_in, ys_out = pad_sos_eos(y, 1, 1)

        # T,B
        tgt_input = ys_in.transpose(1,0)
        tgt_out = ys_out.transpose(1,0)

        src = x.transpose(1,0)

        src_mask, tgt_mask, src_padding_mask, tgt_padding_mask = create_mask(src, tgt_input)

        logits = self.forward(src, tgt_input, src_mask, tgt_mask,src_padding_mask, tgt_padding_mask, src_padding_mask)

        loss = self.criterion(logits.reshape(-1, logits.shape[-1]), tgt_out.reshape(-1))

        return loss


    def inference(self, x):

        self.eval()
        src = x.view(-1, 1)
        num_tokens = src.shape[0]
        src_mask = (torch.zeros(num_tokens, num_tokens)).type(torch.bool).to(DEVICE)
        max_len=num_tokens + 5

        memory = self.encode(src, src_mask)
        ys = x.new_ones(1, 1).fill_(BOS_IDX).type(torch.long)

        for i in range(max_len-1):
            memory = memory.to(DEVICE)
            tgt_mask = (generate_square_subsequent_mask(ys.size(0))
                        .type(torch.bool)).to(DEVICE)
            out = self.decode(ys, memory, tgt_mask)
            out = out.transpose(0, 1)
            prob = self.generator(out[:, -1])
            _, next_word = torch.max(prob, dim=1)
            next_word = next_word.item()

            ys = torch.cat([ys,
                            torch.ones(1, 1).type_as(src.data).fill_(next_word)], dim=0)
            if next_word == EOS_IDX:
                break


        out = ys.squeeze(1).tolist()[1:]

        if out[-1] == 1:
            out = out[:-1]

        return out


from setuptools import setup,find_packages

requirements=[r.strip() for r in open("requirements.txt").readlines()]

setup(
   name='transphone',
   version='1.0.8',
   description='a multilingual g2p/p2g model',
   author='Xinjian Li',
   author_email='xinjianl@cs.cmu.edu',
   url="https://github.com/xinjli/transphone",
   packages=find_packages(),
   install_requires=requirements,
)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from collections import OrderedDict
from typing import Union
import collections
import logging
import os
import re
import traceback
import shutil

import torch
from torch.serialization import default_restore_location
from pathlib import Path


def torch_save(model, path):
    """Save torch model states.

    Args:
        path (str): Model path to be saved.
        model (torch.nn.Module): Torch model.

    """
    path = str(path)
    if hasattr(model, 'module'):
        torch.save(model.module.state_dict(), path)
    else:
        torch.save(model.state_dict(), path)


def torch_load(model, path):
    """Load torch model states.

    Args:
        path (str): Model path or snapshot file path to be loaded.
        model (torch.nn.Module): Torch model.

    """
    model_state_dict = torch.load(str(path), map_location=torch.device('cpu'))

    new_state_dict = OrderedDict()
    for k, v in model_state_dict.items():

        if k.startswith('module.'):
            name = k[7:] # remove `module.`
        else:
            name = k

        new_state_dict[name] = v

    if hasattr(model, 'module'):
        model.module.load_state_dict(new_state_dict)
    else:
        model.load_state_dict(new_state_dict)

    del model_state_dict, new_state_dict


def save_checkpoint(args, trainer, epoch_itr, val_loss):
    from pyspeech.ml.torch import distributed_utils, meters

    prev_best = getattr(save_checkpoint, 'best', val_loss)
    if val_loss is not None:
        best_function = max if args.maximize_best_checkpoint_metric else min
        save_checkpoint.best = best_function(val_loss, prev_best)

    if args.no_save or not distributed_utils.is_master(args):
        return

    def is_better(a, b):
        return a >= b if args.maximize_best_checkpoint_metric else a <= b

    write_timer = meters.StopwatchMeter()
    write_timer.start()

    epoch = epoch_itr.epoch
    end_of_epoch = epoch_itr.end_of_epoch()
    updates = trainer.get_num_updates()

    checkpoint_conds = collections.OrderedDict()
    checkpoint_conds['checkpoint{}.pt'.format(epoch)] = (
        end_of_epoch and not args.no_epoch_checkpoints and
        epoch % args.save_interval == 0
    )
    checkpoint_conds['checkpoint_{}_{}.pt'.format(epoch, updates)] = (
        not end_of_epoch and args.save_interval_updates > 0 and
        updates % args.save_interval_updates == 0
    )
    checkpoint_conds['checkpoint_best.pt'] = (
        val_loss is not None and
        (not hasattr(save_checkpoint, 'best') or is_better(val_loss, save_checkpoint.best))
    )
    checkpoint_conds['checkpoint_last.pt'] = not args.no_last_checkpoints

    extra_state = {
        'train_iterator': epoch_itr.state_dict(),
        'val_loss': val_loss,
    }
    if hasattr(save_checkpoint, 'best'):
        extra_state.update({'best': save_checkpoint.best})

    checkpoints = [os.path.join(args.save_dir, fn) for fn, cond in checkpoint_conds.items() if cond]
    if len(checkpoints) > 0:
        trainer.save_checkpoint(checkpoints[0], extra_state)
        for cp in checkpoints[1:]:
            shutil.copyfile(checkpoints[0], cp)

        write_timer.stop()
        print('| saved checkpoint {} (epoch {} @ {} updates) (writing took {} seconds)'.format(
            checkpoints[0], epoch, updates, write_timer.sum))

    if not end_of_epoch and args.keep_interval_updates > 0:
        # remove old checkpoints; checkpoints are sorted in descending order
        checkpoints = checkpoint_paths(
            args.save_dir, pattern=r'checkpoint_\d+_(\d+)\.pt',
        )
        for old_chk in checkpoints[args.keep_interval_updates:]:
            if os.path.lexists(old_chk):
                os.remove(old_chk)

    if args.keep_last_epochs > 0:
        # remove old epoch checkpoints; checkpoints are sorted in descending order
        checkpoints = checkpoint_paths(
            args.save_dir, pattern=r'checkpoint(\d+)\.pt',
        )
        for old_chk in checkpoints[args.keep_last_epochs:]:
            if os.path.lexists(old_chk):
                os.remove(old_chk)


def load_checkpoint(args, trainer, data_selector=None):
    """Load a checkpoint and restore the training iterator."""
    # only one worker should attempt to create the required dir
    if args.distributed_rank == 0:
        os.makedirs(args.save_dir, exist_ok=True)

    if args.restore_file == 'checkpoint_last.pt':
        checkpoint_path = os.path.join(args.save_dir, 'checkpoint_last.pt')
    else:
        checkpoint_path = args.restore_file

    extra_state = trainer.load_checkpoint(
        checkpoint_path,
        args.reset_optimizer,
        args.reset_lr_scheduler,
        eval(args.optimizer_overrides),
        reset_meters=args.reset_meters,
    )

    if (
        extra_state is not None
        and 'best' in extra_state
        and not args.reset_optimizer
        and not args.reset_meters
    ):
        save_checkpoint.best = extra_state['best']

    if extra_state is not None and not args.reset_dataloader:
        # restore iterator from checkpoint
        itr_state = extra_state['train_iterator']
        epoch_itr = trainer.get_train_iterator(epoch=itr_state['epoch'], load_dataset=True, data_selector=data_selector)
        epoch_itr.load_state_dict(itr_state)
    else:
        epoch_itr = trainer.get_train_iterator(epoch=0, load_dataset=True, data_selector=data_selector)

    trainer.lr_step(epoch_itr.epoch)

    return extra_state, epoch_itr


def load_checkpoint_to_cpu(path, arg_overrides=None):
    """Loads a checkpoint to CPU (with upgrading for backward compatibility)."""
        # if path manager not found, continue with local file.
    state = torch.load(path, map_location=lambda s, l: default_restore_location(s, 'cpu'),)

    args = state['args']
    if arg_overrides is not None:
        for arg_name, arg_val in arg_overrides.items():
            setattr(args, arg_name, arg_val)
    state = _upgrade_state_dict(state)
    return state


def load_model_ensemble(filenames, arg_overrides=None, task=None):
    """Loads an ensemble of models.

    Args:
        filenames (List[str]): checkpoint files to load
        arg_overrides (Dict[str,Any], optional): override model args that
            were used during model training
        task (fairseq.tasks.FairseqTask, optional): task to use for loading
    """
    ensemble, args, _task = load_model_ensemble_and_task(filenames, arg_overrides, task)
    return ensemble, args


def load_model_ensemble_and_task(filenames, arg_overrides=None, task=None):
    from fairseq import tasks

    ensemble = []
    for filename in filenames:
        if not os.path.exists(filename):
            raise IOError('Model file not found: {}'.format(filename))
        state = load_checkpoint_to_cpu(filename, arg_overrides)

        args = state['args']
        if task is None:
            task = tasks.setup_task(args)

        # build model for ensemble
        model = task.build_model(args)
        model.load_state_dict(state['model'], strict=True)
        ensemble.append(model)
    return ensemble, args, task


def checkpoint_paths(path, pattern=r'checkpoint(\d+)\.pt'):
    """Retrieves all checkpoints found in `path` directory.

    Checkpoints are identified by matching filename to the specified pattern. If
    the pattern contains groups, the result will be sorted by the first group in
    descending order.
    """
    pt_regexp = re.compile(pattern)
    files = os.listdir(path)

    entries = []
    for i, f in enumerate(files):
        m = pt_regexp.fullmatch(f)
        if m is not None:
            idx = int(m.group(1)) if len(m.groups()) > 0 else i
            entries.append((idx, m.group(0)))
    return [os.path.join(path, x[1]) for x in sorted(entries, reverse=True)]


def torch_persistent_save(*args, **kwargs):
    for i in range(3):
        try:
            return torch.save(*args, **kwargs)
        except Exception:
            if i == 2:
                logging.error(traceback.format_exc())


def convert_state_dict_type(state_dict, ttype=torch.FloatTensor):
    if isinstance(state_dict, dict):
        cpu_dict = OrderedDict()
        for k, v in state_dict.items():
            cpu_dict[k] = convert_state_dict_type(v)
        return cpu_dict
    elif isinstance(state_dict, list):
        return [convert_state_dict_type(v) for v in state_dict]
    elif torch.is_tensor(state_dict):
        return state_dict.type(ttype)
    else:
        return state_dict


def save_state(
    filename, args, model_state_dict, criterion, optimizer, lr_scheduler,
    num_updates, optim_history=None, extra_state=None,
):
    from fairseq import utils
    if optim_history is None:
        optim_history = []
    if extra_state is None:
        extra_state = {}
    state_dict = {
        'args': args,
        'model': model_state_dict if model_state_dict else {},
        'optimizer_history': optim_history + [
            {
                'criterion_name': criterion.__class__.__name__,
                'optimizer_name': optimizer.__class__.__name__,
                'lr_scheduler_state': lr_scheduler.state_dict(),
                'num_updates': num_updates,
            }
        ],
        'extra_state': extra_state,
    }
    if utils.has_parameters(criterion):
        state_dict['criterion'] = criterion.state_dict()
    if not args.no_save_optimizer_state:
        state_dict['last_optimizer_state'] = convert_state_dict_type(optimizer.state_dict())

    torch_persistent_save(state_dict, filename)


def _upgrade_state_dict(state):
    """Helper for upgrading old model checkpoints."""
    from fairseq import models, registry, tasks

    # add optimizer_history
    if 'optimizer_history' not in state:
        state['optimizer_history'] = [
            {
                'criterion_name': 'CrossEntropyCriterion',
                'best_loss': state['best_loss'],
            },
        ]
        state['last_optimizer_state'] = state['optimizer']
        del state['optimizer']
        del state['best_loss']
    # move extra_state into sub-dictionary
    if 'epoch' in state and 'extra_state' not in state:
        state['extra_state'] = {
            'epoch': state['epoch'],
            'batch_offset': state['batch_offset'],
            'val_loss': state['val_loss'],
        }
        del state['epoch']
        del state['batch_offset']
        del state['val_loss']
    # reduce optimizer history's memory usage (only keep the last state)
    if 'optimizer' in state['optimizer_history'][-1]:
        state['last_optimizer_state'] = state['optimizer_history'][-1]['optimizer']
        for optim_hist in state['optimizer_history']:
            del optim_hist['optimizer']
    # record the optimizer class name
    if 'optimizer_name' not in state['optimizer_history'][-1]:
        state['optimizer_history'][-1]['optimizer_name'] = 'FairseqNAG'
    # move best_loss into lr_scheduler_state
    if 'lr_scheduler_state' not in state['optimizer_history'][-1]:
        state['optimizer_history'][-1]['lr_scheduler_state'] = {
            'best': state['optimizer_history'][-1]['best_loss'],
        }
        del state['optimizer_history'][-1]['best_loss']
    # keep track of number of updates
    if 'num_updates' not in state['optimizer_history'][-1]:
        state['optimizer_history'][-1]['num_updates'] = 0
    # old model checkpoints may not have separate source/target positions
    if hasattr(state['args'], 'max_positions') and not hasattr(state['args'], 'max_source_positions'):
        state['args'].max_source_positions = state['args'].max_positions
        state['args'].max_target_positions = state['args'].max_positions
    # use stateful training data iterator
    if 'train_iterator' not in state['extra_state']:
        state['extra_state']['train_iterator'] = {
            'epoch': state['extra_state']['epoch'],
            'iterations_in_epoch': state['extra_state'].get('batch_offset', 0),
        }
    # default to translation task
    if not hasattr(state['args'], 'task'):
        state['args'].task = 'translation'

    # set any missing default values in the task, model or other registries
    registry.set_defaults(state['args'], tasks.TASK_REGISTRY[state['args'].task])
    registry.set_defaults(state['args'], models.ARCH_MODEL_REGISTRY[state['args'].arch])
    for registry_name, REGISTRY in registry.REGISTRIES.items():
        choice = getattr(state['args'], registry_name, None)
        if choice is not None:
            cls = REGISTRY['registry'][choice]
            registry.set_defaults(state['args'], cls)

    return state


def verify_checkpoint_directory(model_path: Path) -> None:

    if not model_path.exists():
        model_path.mkdir(exist_ok=True, parents=True)

    temp_file_path = model_path / 'dummy'
    try:
        with open(str(temp_file_path), 'w'):
            pass
    except OSError as e:
        (f'| Unable to access checkpoint save directory: {model_path}')
        raise e
    else:
        os.remove(temp_file_path)



import torch
import torch.nn as nn
from transphone.data.utils import pad_sos_eos
import torch.nn.functional as F
import math
import numpy as np

class Encoder(nn.Module):

    def __init__(self, vocab_size, hidden_size, layer_size):
        super().__init__()

        self.embed = nn.Embedding(vocab_size, hidden_size)
        self.rnn = nn.LSTM(hidden_size, hidden_size, layer_size, batch_first=True)
        #self.linear = nn.Linear(2*hidden_size, hidden_size)

    def forward(self, input_tensor):
        embed = self.embed(input_tensor)
        output, (hidden, _) = self.rnn(embed)
        #output = self.linear(output)

        return output, hidden


class Decoder(nn.Module):

    def __init__(self, vocab_size, hidden_size):
        super().__init__()
        self.embed = nn.Embedding(vocab_size, hidden_size)
        self.rnn = nn.LSTM(hidden_size, hidden_size, 1, batch_first=True)
        self.linear = nn.Linear(hidden_size, vocab_size)
        self.logsoftmax = nn.LogSoftmax(dim=2)

    def forward(self, input_tensor, prev_hidden, prev_cell):
        # input: [B]

        embed = self.embed(input_tensor)
        output, (hidden, cell) = self.rnn(embed, (prev_hidden, prev_cell))
        #print('after rnn', output)
        output = self.linear(output)
        #print('after linear:', output)
        output = self.logsoftmax(output)
        #print('after softmax', output)

        return output, hidden, cell


class AttentionDecoder(nn.Module):

    def __init__(self, vocab_size, hidden_size):
        super().__init__()
        self.embed = nn.Embedding(vocab_size, hidden_size)
        self.rnn = nn.LSTM(2*hidden_size, hidden_size, 1, batch_first=True)
        self.linear = nn.Linear(hidden_size, vocab_size)
        self.linearQ = nn.Linear(2*hidden_size, hidden_size)
        self.logsoftmax = nn.LogSoftmax(dim=2)
        self.hidden_size = hidden_size

    def forward(self, input_tensor, encoder_vector, encoder_mask, prev_hidden, prev_cell):
        # input: [B, 1]
        # encoder: [B,T,H]
        # encoder mask: (B,T)
        # prev_hidden: (1,B,H)
        # prev_cell: (1,B,H)

        # [B,1,H]
        embed = self.embed(input_tensor).squeeze(1)

        # [B,2H] -> [B,H,1]
        Q = self.linearQ(torch.cat([embed, prev_hidden.squeeze(0)], dim=1)).unsqueeze(-1)
        K = encoder_vector

        # (B,T)
        unnormed_weights = torch.bmm(K, Q).squeeze(2)/math.sqrt(self.hidden_size)

        # masking (B,T)
        masked_weights = unnormed_weights.masked_fill(~encoder_mask, -np.inf)

        # [B,T,H] [B,H,1] -> [B,T,1] -> [B,T]
        attn_weights = F.softmax(masked_weights, dim=1)

        # [B,H,T]x[B,T,1] -> [B,H,1] -> [B,H]
        attn_applied = torch.bmm(encoder_vector.transpose(1,2), attn_weights.unsqueeze(-1)).squeeze(2)

        # (B,2H) -> (B,1,2H)
        lstm_input = torch.cat([attn_applied, embed], dim=1).unsqueeze(1)

        # [B,1,2H]
        output, (hidden, cell) = self.rnn(lstm_input, (prev_hidden, prev_cell))

        #print('after rnn', output)
        output = self.linear(output)
        #print('after linear:', output)
        output = self.logsoftmax(output)
        #print('after softmax', output)

        return output, hidden, cell, attn_weights


class G2P(nn.Module):

    def __init__(self):
        super().__init__()

        self.hidden_size = 256
        self.vocab_size = 200
        self.encoder = Encoder(self.vocab_size, self.hidden_size, 1)
        self.decoder = Decoder(self.vocab_size, self.hidden_size)

        self.criterion = nn.NLLLoss(ignore_index=0)

    def train_step(self, x, y):

        self.train()
        batch_size = x.shape[0]

        output, prev_hidden = self.encoder(x)
        prev_cell = x.new_zeros(1, batch_size, self.hidden_size, dtype=torch.float)

        ys_in, ys_out = pad_sos_eos(y, 1, 1)
        ys_in = ys_in.transpose(1,0)
        ys_out = ys_out.transpose(1,0)

        loss = 0

        for i in range(len(ys_in)):
            y_in = ys_in[i].view(batch_size, 1)
            y_out = ys_out[i]

            output, prev_hidden, prev_cell = self.decoder(y_in, prev_hidden, prev_cell)

            output = output.squeeze()
            #print('----')
            #print(y_out.shape)
            #print(output.shape)
            #print('y_out', y_out)
            #print('output', output)
            loss += self.criterion(output, y_out)

        return loss

    def inference(self, x):

        self.eval()

        #x = torch.LongTensor([[52, 74, 57, 72, 63,  0,  0,  0,  0,  0,  0,  0]])

        batch_size = 1

        output, prev_hidden = self.encoder(x)
        prev_cell = prev_hidden.new_zeros((1, batch_size, self.hidden_size), dtype=torch.float)

        #print(prev_hidden)

        #y_in = torch.LongTensor([[1]])
        y_out = []

        w = 1

        while(True):
            y_in = x.new([[w]])
            output, prev_hidden, prev_cell = self.decoder(y_in, prev_hidden, prev_cell)
            output = output.squeeze()
            #print(output)
            w = output.data.topk(1)[1].item()
            y_out.append(w)

            if w == 1 or len(y_out)>16:
                break

        return y_out


class AttnG2P(nn.Module):

    def __init__(self):
        super().__init__()

        self.hidden_size = 512
        self.vocab_size = 200
        self.encoder = Encoder(self.vocab_size, self.hidden_size, 2)
        self.decoder = AttentionDecoder(self.vocab_size, self.hidden_size)

        self.criterion = nn.NLLLoss(ignore_index=0)

    def train_step(self, x, y):

        self.train()
        batch_size = x.shape[0]
        encoder_mask = (x != 0)

        encoder_output, _ = self.encoder(x)
        prev_cell = x.new_zeros(1, batch_size, self.hidden_size, dtype=torch.float)
        prev_hidden = x.new_zeros(1, batch_size, self.hidden_size, dtype=torch.float)

        ys_in, ys_out = pad_sos_eos(y, 1, 1)
        ys_in = ys_in.transpose(1,0)
        ys_out = ys_out.transpose(1,0)

        loss = 0

        for i in range(len(ys_in)):
            y_in = ys_in[i].view(batch_size, 1)
            y_out = ys_out[i]

            output, prev_hidden, prev_cell, _ = self.decoder(y_in, encoder_output, encoder_mask, prev_hidden, prev_cell)

            output = output.squeeze()
            #print('----')
            #print(y_out.shape)
            #print(output.shape)
            #print('y_out', y_out)
            #print('output', output)
            loss += self.criterion(output, y_out)

        return loss

    def inference_with_attention(self, x):

        self.eval()
        encoder_mask = (x != 0)

        #x = torch.LongTensor([[52, 74, 57, 72, 63,  0,  0,  0,  0,  0,  0,  0]])

        batch_size = 1

        encoder_output, _ = self.encoder(x)
        prev_cell = x.new_zeros((1, batch_size, self.hidden_size), dtype=torch.float)
        prev_hidden = x.new_zeros(1, batch_size, self.hidden_size, dtype=torch.float)

        #print(prev_hidden)

        #y_in = torch.LongTensor([[1]])
        y_out = []

        w = 1

        weights = []

        while(True):
            y_in = x.new([[w]])
            output, prev_hidden, prev_cell, attn_weights = self.decoder(y_in, encoder_output, encoder_mask, prev_hidden, prev_cell)
            output = output.squeeze()
            w = output.data.topk(1)[1].item()

            weights.append(attn_weights.unsqueeze(-1))
            if w == 1 or len(y_out)>16:
                break

            y_out.append(w)

        weights = torch.cat(weights, dim=2)

        return y_out, weights



    def inference(self, x):

        self.eval()

        #x = torch.LongTensor([[52, 74, 57, 72, 63,  0,  0,  0,  0,  0,  0,  0]])

        batch_size = 1

        encoder_output, prev_hidden = self.encoder(x)
        prev_cell = prev_hidden.new_zeros((1, batch_size, self.hidden_size), dtype=torch.float)

        #print(prev_hidden)

        #y_in = torch.LongTensor([[1]])
        y_out = []

        w = 1

        while(True):
            y_in = x.new([[w]])
            output, prev_hidden, prev_cell = self.decoder(y_in, encoder_output, prev_hidden, prev_cell)
            output = output.squeeze()
            #print(output)
            w = output.data.topk(1)[1].item()
            y_out.append(w)

            if w == 1 or len(y_out)>16:
                break

        return y_out